/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Louis-Marie Traonouez <louis-marie.traonouez@inria.fr>
*/
package preprocessor;

import java.util.HashSet;

/**
 *
 * @author ltraonou
 */
public class AssignStatement extends Statement {
    private String var;
    private String expr;

    public AssignStatement() {
    	usedVariables = new HashSet<Variable>();
    	nbStatements = 1;
    }
    
    public AssignStatement(String v, String e) {
    	usedVariables = new HashSet<Variable>();
		nbStatements = 1;
        var = v;
        expr = e;
    }
    
    public void setVariable(String v) {
    	var = v;
    }
    
    public void setExpression(String e) {
    	expr = e;
    }
   
    public Statement replace(String oldv, String newv) { 
        String newVar = var.replace("("+oldv+")","("+newv+")");
        newVar = newVar.replace("("+oldv+"[","("+newv+"[");
        String newExpr = expr.replace("("+oldv+")","("+newv+")");
        newExpr = newExpr.replace("("+oldv+"[","("+newv+"[");
        AssignStatement ret = new AssignStatement(newVar, newExpr);
        ret.addUsedVariable(getUsedVariables());
        for (Variable v : ret.getUsedVariables()) {
        	v.setLastUsed(ret);
        }
        return ret;
    }

    public String toString(int line) {
        return "assign " + var + ":=" + expr + ";\n";
    }
}
