/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Louis-Marie Traonouez <louis-marie.traonouez@inria.fr>
*/

package preprocessor;

import java.io.PrintWriter;
import java.math.BigDecimal;
import java.util.HashMap;
import org.antlr.runtime.*;

/**
 *
 * @author ltraonou
 */
public class Preprocessor {
    public static void preprocess(String filename, HashMap<String,BigDecimal> constants, boolean free) throws Exception {
	Programme prog = new Programme(constants);
        PreprocessorLexer lexer = new PreprocessorLexer(new ANTLRFileStream(filename));
        CommonTokenStream tokens = new CommonTokenStream(lexer);
        PreprocessorParser parser = new PreprocessorParser(tokens);
        parser.prog_grammar(prog);
	prog.checkConstants();
        
        filename = filename.concat(".pp");
        if (free)
	  prog.freeVariables();
        prog.print(new PrintWriter(filename));
    }
    
    public static void main(String[] args) throws Exception {
        preprocess(args[0],null,true);
    }
}
