/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Louis-Marie Traonouez <louis-marie.traonouez@inria.fr>
*/

package preprocessor;

import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.Iterator;
import java.util.LinkedList;

/**
 *
 * @author ltraonou
 */
public class PrivateVariable extends Variable {
   
    private boolean publicity; // true for PRIVATE, false for SECRET
    private LinkedList<BigInteger> intRange;
    
    public PrivateVariable(int t) {
        intType = t;
        minInt = BigInteger.ZERO;
        maxInt = new BigInteger("2").pow(t).subtract(BigInteger.ONE);
        arrays = new LinkedList<ArrayDimension>();
        publicity = true;
        intRange = new LinkedList<BigInteger>();
        lastUsed = null;
    }

    public void setPublicity(boolean b) {
        publicity = b;
    }
    
    // 1 for public
    // 2 for observable
    // -1 for private
    // -2 for secret
    public int getPublicity() {
        if (publicity) {
            return -1;
        }
        else {
            return -2;
        }
    }
      
    public int updateIntRange(LinkedList<BigInteger> l) {
        if (l.getFirst().compareTo(minInt) >= 0 && l.getLast().compareTo(maxInt) <= 0) {
            intRange = l;
            return 0;
        }
        else {
            return 2;
        }
    }
    
     public int print(PrintWriter out) {
        String decl = "";
        if (publicity) {
            decl = "private";
        }
        else {
            decl = "secret";
        }
        decl = decl.concat(" int"+intType+" "+name);
        
        String rg = "";
        for (Iterator<BigInteger> it = intRange.iterator(); it.hasNext();) {
             BigInteger valmin = it.next();
             BigInteger valmax = it.next();
             rg = rg.concat("["+valmin.toString()+","+valmax.toString()+"]");
        }

        int line=0;
        if (getDimensions() > 0) {
            for (Iterator<String> it = generateIndices().iterator(); it.hasNext();) {
                String s = it.next();
                String declit =  decl.concat(s);

                if (!rg.isEmpty()) {
                    declit = declit.concat(":=" + rg);
                }
                declit = declit.concat(";");
                out.println(declit);
                line++;
            }
        }
        else {
            if (!rg.isEmpty()) {
                decl = decl.concat(":=" + rg);
            }
            decl = decl.concat(";");
            out.println(decl);
            line++;
        }
        return line;
    }
}
