/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package quail;

import java.io.FileNotFoundException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 *
 * @author fbio
 */
public class ConcurrentExplorer{
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();

    private int numthreads = 2;
   

   

    ConcurrentExplorer(int threads) {
        numthreads = threads;
    }
    
    
    void explore(String filename) {
        ArrayList<String> codelines = null;

        long starttime = System.currentTimeMillis();
        CodeReader reader = new CodeReader(filename);
       
        try{
            codelines = reader.readLines();
        }catch (FileNotFoundException fnf)
        {
            log("Error while processing input file: file " + filename + " not found",1);
            System.exit(1);
            
        }
        //initializing
        AtomicInteger count = new AtomicInteger(1);
        long keepAliveTime = 1000;
        
        MarkovChainState firststate = new MarkovChainState("current",contexter.getDecimalOne());
        MarkovStatePublicVariable programcounter = new MarkovStatePublicVariable("pc",BigInteger.ONE,Integer.valueOf(32));
        firststate.addVariable(programcounter);
        
        PriorDistributionManager priordistman = new PriorDistributionManager(codelines);
                
        BigDecimal priorentropy = priordistman.computePriorShannonEntropy();
        
        ConcurrentSimplePosteriorDistributionManager postdistman = new ConcurrentSimplePosteriorDistributionManager();
        
        log("Starting state space exploration at time "+(System.currentTimeMillis()-starttime)+" milliseconds using "+numthreads+" threads",2);
        //ThreadPoolExecutor tpoolexplorer = Executors.newCachedThreadPool();
        ThreadPoolExecutor tpoolexplorer = new ThreadPoolExecutor(numthreads, numthreads, keepAliveTime, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>());
        tpoolexplorer.submit(new RunnableSingleExplorer(firststate,codelines,postdistman,tpoolexplorer,count));
        
        try {
            tpoolexplorer.awaitTermination(1, TimeUnit.HOURS);
        } catch (InterruptedException ex) {
            log("ERROR: Explorer interrupted: "+ex.getLocalizedMessage()+". Terminating",1);
            System.exit(1);
        }
        /*
        
        boolean done = false;
        int tracesexplored = 1;
        while (!done){
            log("\n------running new eploration ---",3);
            MarkovChainState finalstate = expengine.exploreOnce();
            if (finalstate == null){
                done = true;
                break;
            }else{
                postdistman.inferInformation(finalstate);
            }
            tracesexplored++;
            if (tracesexplored%pace==0){
                BigDecimal percentcompleted = postdistman.getTotalProbability().scaleByPowerOfTen(2);
                log(percentcompleted.toEngineeringString()+"% of exploration completed at time "+(System.currentTimeMillis()-starttime)+" milliseconds",2);
            }
                           
            if (!fast){
                   try {
                    Thread.sleep(1);
                } catch (InterruptedException ex) {
                    java.util.logging.Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, null, ex);
                } 
            }
        }*/
        log("State space exploration terminated  at time "+(System.currentTimeMillis()-starttime)+" milliseconds after exploring "+tpoolexplorer.getCompletedTaskCount()+" traces",2);
        
        log("Output distributions:\n"+postdistman.getDistributionOnOutputs().printNice(),3);
        log("Examining outputs...",2);
        

        BigDecimal postentropy = postdistman.computePosteriorShannonEntropy();
        log("The prior entropy of the secret is "+priorentropy,2);
        log("The posterior entropy is "+postentropy,2);
        
        
        BigDecimal leakage = priorentropy.subtract(postentropy);
        if (logger.getVerbosity()>0){
            log("The leakage is "+leakage+" bits",1);
        }else{
            log(leakage,0);
        }
//        double interppostentropy = getInterpolatedValue(interpx,interpy,1.0);
//        log("The interpolated posterior entropy estimate is "+interppostentropy,2);
//        BigDecimal interpleakage = secretbitsize.subtract(new BigDecimal(interppostentropy));
//        log("The interpolated leakage is "+interpleakage+" bits",2);
        
        log("Time for exploration: "+(System.currentTimeMillis()-starttime)+" milliseconds",2);
        
    }
    
    
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }
    
    
}
