/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package quail;

import java.math.BigDecimal;
import java.util.HashMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;

/**
 *
 * @author fbio
 */
class ConcurrentPosteriorDistributionManager {
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();

    private final ConcurrentHashMap<String,MultiIntervalProbabilityDistribution> outputs = new ConcurrentHashMap<>();
    private final ConcurrentWeightedProbabilityDistribution outputdist = new ConcurrentWeightedProbabilityDistribution();
    IntervalAssignments initialsecretintervals;
    private final ReentrantLock lock = new ReentrantLock();

    ConcurrentPosteriorDistributionManager(HashMap<String, Interval> initsecint) {
       if (initsecint == null){
           initialsecretintervals = null;
       }else{
           initialsecretintervals = new IntervalAssignments(initsecint);
       }
    }

    void inferInformation(MarkovChainState finalstate) {
        //BigDecimal entropy = new BigDecimal(0);
        //BigInteger newsecretsize = new BigInteger("0");
        VariableList secrets = new VariableList();
        VariableList observables = new VariableList();
        VariableList vars = finalstate.getVariables();


        // TEMPORARY - TESTING ONLY
       // logger.setVerbosity(3);

        for (MarkovStateVariable var : vars){
            if (var.isPrivate()){

                MarkovStatePrivateVariable privvar =(MarkovStatePrivateVariable)var;
                if (privvar.isSecret()){
                    secrets.add(privvar);
                }

                //newsecretsize = newsecretsize.add(asecret.getSize());
            }else{
                MarkovStatePublicVariable pubvar =(MarkovStatePublicVariable)var;
                if (pubvar.isObservable()){
                    observables.add(pubvar);
                }
            }
        }
        log("",3);
        log("Observation: "+observables.getIdentifier()+"\nSecret: "+secrets.getIdentifier(),3);

        // use the name and value of the observables as key. the identifier is a string like "o=17;"
        String key = observables.getIdentifier();
        BigDecimal stateprob = finalstate.getProbability();
        String varidentifier = vars.getIdentifier();

        log("Updating probability distribution after observing "+key+" with variables "+varidentifier+" and probability "+stateprob.toPlainString(),3);

        // outputdist is the probability distributions over all observed outputs
        outputdist.update(key, varidentifier, stateprob);
        log("The new distribution on the observables is:\n"+outputdist.printNice(),3);
        log("with total probability "+outputdist.getTotalProbability(),3);

        log("Updating conditional posterior distribution for "+key,3);
        lock.lock();
// outputs maps each possible output to the corresponding posterior distributions over the secrets
        if (outputs.containsKey(key)){ //the output has already appeared; update the corresponding prob dists with the new information about the secrets
            lock.unlock();
            MultiIntervalProbabilityDistribution mapsecrets =(MultiIntervalProbabilityDistribution) outputs.get(key);
            log("Locking posterior distribution of output "+key,3);
            synchronized(mapsecrets){
                log("Old distribution on the secret if the observable is "+key+":\n"+mapsecrets.printNice(),3);
                mapsecrets.update(secrets,varidentifier,stateprob);
                log("New distribution on the secret if the observable is "+key+":\n"+mapsecrets.printNice(),3);
            }
            log("Unlocking posterior distribution of output "+key,3);

        }else{ // the output has been observed for the first time; create a new prob dist
            MultiIntervalProbabilityDistribution newmap = new MultiIntervalProbabilityDistribution();
            newmap.initialize(secrets,varidentifier,stateprob);
            outputs.put(key, newmap); // maps this output to the new prob dist
            lock.unlock();
            log("Created new distributions for output "+key,3);
        }


    }

    BigDecimal getTotalProbability() {
        return outputdist.getTotalProbability();
    }

    ConcurrentWeightedProbabilityDistribution getDistributionOnOutputs() {
        return outputdist;
    }

    private static void log(Object aObject, int level){

        logger.log(String.valueOf(aObject),level);
    }

    BigDecimal computePosteriorShannonEntropy() {
        BigDecimal toreturn = contexter.getDecimalZero();
        //outputdist.normalize();
        BigDecimal totobsprob = outputdist.getTotalProbability();
        boolean willpad = false;
        log("I observed "+outputdist.size()+" outputs for a total observed probability of "+totobsprob.toPlainString(),2);
        if (totobsprob.compareTo(contexter.getDecimalOne())<0 && initialsecretintervals!=null){
            willpad = true;
            log("Padding will be applied.",3);
        }


        for (String aout : outputdist.keySet()){
            BigDecimal prob = outputdist.getProbability(aout);
            log("Output "+aout+" has probability "+prob+".\nIts conditional probability distribution is",3);
            BigDecimal entropy = contexter.getDecimalZero();
            MultiIntervalProbabilityDistribution dist = outputs.get(aout);
            if (willpad){
                log("Posterior distribution on the secrets before padding and normalizing:",3);
            }else{
                log("Posterior distribution on the secrets before normalizing:",3);
            }
            log(dist.printNice(),3);

            if (willpad){
                //pad
                dist.pad(initialsecretintervals);
            }
            //normalize
            dist.normalize();

            if (willpad){
                log("Posterior distribution on the secrets after padding and normalizing:",3);
            }else{
                log("Posterior distribution on the secrets after normalizing:",3);
            }
            log(dist.printNice(),3);

            //get entropy
            log(dist.printNice(),3);
            entropy = entropy.add(dist.getEntropy(),contexter.getContext());
            log("Its conditional entropy is "+entropy,3);


            BigDecimal contr = prob.multiply(entropy, contexter.getContext());
            log("Output "+aout+" contributes "+contr+" bits to the posterior entropy estimate\n",3);
            toreturn = toreturn.add(contr);
        }
        // assume that the unobserved part of the system has the same leakage as the observed part
        // totobsprob = totobsprob.multiply(new BigDecimal(100));
        log("Conditional entropy after observing "+(totobsprob.multiply(new BigDecimal(100))).toEngineeringString()+"% of the probability space is "+toreturn.toEngineeringString(),2);
        toreturn = toreturn.divide(totobsprob, contexter.getContext());

        return toreturn;
    }

}
