/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package quail;

import java.io.FileNotFoundException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;

/**
 *
 * @author fbio
 */
public class ConcurrentSimulator {
    private final int iterations;
    private int pace = 500;
    private final BigDecimal treshold;

    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
    private final long timelimit;
    private int numthreads = 2;
    
    
    //private Interval secretinterval;
    

    
    public ConcurrentSimulator(int iter, BigDecimal tresh, long tlimit, int threads){
        iterations = iter;
        treshold = tresh;
        numthreads = threads;
        timelimit = tlimit;
    }
    
    public void simulate(String filename){
       
        ArrayList<String> codelines = null;
        
        // these are the data arrays for the interpolation. we initialize them to -1 so we can recognize the values we add later
        /*double[] interpx = new double[iterations/pace];
        double[] interpy = new double[iterations/pace];
        int interpcounter = 0;
        for (int i=0;i<iterations/pace;i++){
            interpx[i] = -1;
            interpy[i] = -1;
        }*/
        
        
        
        long time = System.currentTimeMillis();
        CodeReader reader = new CodeReader(filename);
        boolean allexplored = false;
        boolean timeout = false;
        boolean alliterations = false;
        
        try{
            codelines = reader.readLines();
        }catch (FileNotFoundException fnf)
        {
            log("Error while processing input file: file " + filename + " not found",1);
            System.exit(1);
            
        }
        //initializing
        long keepAliveTime = 1000;
        
        
        
        PriorDistributionManager priordistman = new PriorDistributionManager(codelines);
        
        
        
        
        BigDecimal priorentropy = priordistman.computePriorShannonEntropy();
        if (pace>=iterations) {
            pace=0;
        }
        ConcurrentPosteriorDistributionManager postdistman = new ConcurrentPosteriorDistributionManager(priordistman.getInitialSecretIntervals());
        ThreadPoolExecutor tpoolsimulator = new ThreadPoolExecutor(numthreads, numthreads, keepAliveTime, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>());
        
        long simtime = System.currentTimeMillis();
        log("Scheduling "+iterations+" simulations at time "+(simtime-time)+" milliseconds using "+numthreads+" threads",2);
        for (int i=1; i<=iterations; i++){
                tpoolsimulator.submit(new RunnableSingleSimulator(codelines,postdistman));
             
            /*if (allexplored){
                log("In "+i+" simulations we have reached the treshold of "+treshold.toEngineeringString()+", no further simulation needed.",2);
                break;
            }
            
            if (timeout){
                log("In "+i+" simulations we have reached the time limit of "+timelimit+" milliseconds, moving to leakage computation.",2);
                break;
            }*/
            
        }
        //log(iterations+" simulation scheduled",2);
        //tpoolsimulator.shutdown();
        
        while (!allexplored && !timeout && !alliterations){
            try {
                Thread.sleep(1000);
            } catch (InterruptedException ex) {
                java.util.logging.Logger.getLogger(ConcurrentSimulator.class.getName()).log(Level.SEVERE, null, ex);
            }
            BigDecimal totprob = postdistman.getTotalProbability();
        
            long currtime = System.currentTimeMillis();
            if (totprob.compareTo(treshold)>=0){
                allexplored = true;
            }
            if ((currtime-time)>timelimit && timelimit!=0){
                timeout = true;
            }
            long completedsim = tpoolsimulator.getCompletedTaskCount();
            if (completedsim>=iterations){
                alliterations = true;
            }
            log(completedsim+"/"+iterations+" simulations completed at time "+(currtime-time)+" milliseconds, total probability "+totprob.toEngineeringString(),2);
        }
        tpoolsimulator.shutdownNow();
        try {
            tpoolsimulator.awaitTermination(10, TimeUnit.SECONDS);
        } catch (InterruptedException ex) {
            java.util.logging.Logger.getLogger(ConcurrentSimulator.class.getName()).log(Level.SEVERE, null, ex);
        }
        if (allexplored){
            log("We have reached the treshold of "+treshold.toEngineeringString()+", no further simulation needed.",2);
        }
        if (timeout){
            log("We have reached the time limit of "+timelimit+" milliseconds, moving to leakage computation.",2);
        }   
        if (alliterations){
            log(iterations+" simulations completed, moving to leakage computation.",2);
        }  
        
        //log("\n"+iterations+" simulations terminated at time "+(System.currentTimeMillis()-time)+" milliseconds",2);
        log("Output distributions:\n"+postdistman.getDistributionOnOutputs().printNice(),3);
        log("Examining outputs:\n",3);
        

        BigDecimal postentropy = postdistman.computePosteriorShannonEntropy();
        log("The prior entropy of the secret is "+priorentropy,2);
        log("The posterior entropy estimate is "+postentropy,2);
        
        
        BigDecimal estleakage = priorentropy.subtract(postentropy);
        if (logger.getVerbosity()>0){
            log("The estimated leakage is "+estleakage+" bits",1);
        }else{
            log(estleakage,0);
        }
//        double interppostentropy = getInterpolatedValue(interpx,interpy,1.0);
//        log("The interpolated posterior entropy estimate is "+interppostentropy,2);
//        BigDecimal interpleakage = secretbitsize.subtract(new BigDecimal(interppostentropy));
//        log("The interpolated leakage is "+interpleakage+" bits",2);
        
        log("Time for extimation: "+(System.currentTimeMillis()-time)+" milliseconds",2);
        
        
      
        
        
    }
    
    
    
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }
}
