/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package quail;

import java.math.BigDecimal;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 *
 * @author fbio
 */
class ConcurrentWeightedProbabilityDistribution extends ConcurrentHashMap<String,HashFreqProb> implements ProbabilityDistribution{

    protected static final Contexter contexter = Contexter.getInstance();
    protected static final Logger logger = Logger.getInstance();
    
    public synchronized void update (String key, String variables, BigDecimal probability){
        if (this.containsKey(key)){
            HashFreqProb internal = this.get(key);
            //synchronized(internal){
                    if (internal.containsKey(variables)){

                    internal.updateProbability(variables,probability);
                }else{
                    internal.create(variables, probability);
                }
            //}
            
            
        }else{
            HashFreqProb internal = new HashFreqProb(variables, probability);
            this.put(key, internal);
        }
    }
  
    @Override
    public BigDecimal getProbability(String key) {
        return this.get(key).getTotalProbability();
    }
    
    public Integer getFrequency(String key) {
        return this.get(key).getTotalFrequency();
    }

    @Override
    public BigDecimal getEntropy() {
        BigDecimal entropy = new BigDecimal(0,contexter.getContext());
        Set<String> keys = this.keySet();
        for (String key : keys){
            BigDecimal prob = this.getProbability(key);
            BigDecimal localent = prob.multiply(log2(prob), contexter.getContext());
            log("Local entropy: "+localent,4);
            entropy = entropy.subtract(localent, contexter.getContext());
        }
        
        return entropy;
    }
    protected static void log(Object aObject, int level){
        logger.log(String.valueOf(aObject),level);
    }

    
    protected BigDecimal log2(BigDecimal value) {
        if (value.compareTo(contexter.getDecimalOne())==0){
            return contexter.getDecimalZero();
        }
        log("Computing the logarithm in base 2 of "+value.toPlainString()+" with scale "+value.scale(),4);
        BigDecimal toreturn = BigFunctions.ln(value,contexter.getContext().getPrecision()+10);
        log("The natural logarithm is "+toreturn,4);
        toreturn = toreturn.divide(BigFunctions.ln(new BigDecimal("2",contexter.getContext()),contexter.getContext().getPrecision()),contexter.getContext());
        log("The logarithm in base 2 is "+toreturn,4);
        
        return toreturn;
    }

    @Override
    public String printNice(){
        String toreturn = "";
        Set<String> keys = this.keySet();
        for (String key : keys){
            toreturn += key+" with value "+this.getFrequency(key)+" and probability "+this.getProbability(key).toPlainString()+"\n";
        }
        return toreturn;
    }

    public void normalize() {
        this.normalizeTo(contexter.getDecimalOne());

    }
    
    public BigDecimal getTotalProbability() {
        BigDecimal toreturn = contexter.getDecimalZero();
        Set<String> keys = this.keySet();
        for (String key : keys){
            toreturn = toreturn.add(this.getProbability(key), contexter.getContext());
        }
        log("The total probability for the distribution is "+toreturn.toPlainString(),4);
        return toreturn;
    }
    
    public Integer getTotalFrequency() {
        Integer toreturn = 0;
        Set<String> keys = this.keySet();
        for (String key : keys){
            toreturn += this.getFrequency(key);
        }
        log("The total freqency for the distribution is "+toreturn,4);
        return toreturn;
    }
    /*
    public BigDecimal getPaddedEntropy(BigDecimal assumedprob){
        BigDecimal totprob = this.getTotalProbability();
        BigDecimal entropy = this.getEntropy();
        if (totprob.compareTo(contexter.getDecimalOne())==0){
            return entropy;
        }else{
            return entropy.divide(totprob, contexter.getContext());
        }
        
    }*/
    
    public void normalizeTo(BigDecimal prob) {
        BigDecimal currprob = this.getTotalProbability();
        if (currprob.compareTo(prob)!=0){
            BigDecimal multiplier = prob.divide(currprob, contexter.getContext());
            for (HashFreqProb map : this.values()){
                this.multiplyInternalProbability(map, multiplier);
            }
        }
    }
    
    protected void updateInternalProbability(HashFreqProb internal, String variables, BigDecimal prob) {
        
        FrequencyAndProbability freqprob = internal.get(variables);
        if (freqprob!=null){
                freqprob.updateProbability(prob);
                log("Interval already exists with the same variables, updated its probability and increased its frequency by 1",4);
                
            }else{
                 internal.put(variables, new FrequencyAndProbability(1,prob));
                 log("Interval already exists but not with these variables, put it in with frequency 1 and probability "+prob.toPlainString(),4);
            }
        
    }

    protected void multiplyInternalProbability(HashFreqProb internal, BigDecimal multiplier) {
        log("Multiplying all internal probabilities by "+multiplier.toEngineeringString(),4);
        for (FrequencyAndProbability freqprob : internal.values()){
            BigDecimal newprob = freqprob.getProbability().multiply(multiplier, contexter.getContext());
            log("New probability: "+newprob.toEngineeringString(),4);
            freqprob.setProbability(newprob);
        }
    }
  
    
}
