/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.math.BigDecimal;
import java.math.MathContext;

/**
 *
 * @author fbio
 */
public class Contexter {
    
    private static Contexter instance;
    private static MathContext context;
    private BigDecimal decimalOne;
    private BigDecimal decimalZero;
    
    private Contexter(){
        context = MathContext.UNLIMITED;
        this.computeConstants();
    }
    
    private void computeConstants(){
    	String zeroes = "";
    	if (context.getPrecision()>1){
    		zeroes = ".";
    		for (int i=0; i<context.getPrecision(); i++){
    			zeroes += "0";
    		}
    			
    	}
        this.decimalOne = new BigDecimal("1"+zeroes,this.context);
        this.decimalZero = new BigDecimal("0"+zeroes,this.context);
    }
    
    public static Contexter getInstance(){
        if (instance == null){
            instance = new Contexter();
        }
        return instance;
    }

    public MathContext getContext() {
        return context;
    }

    public void setContext(MathContext context) {
        Contexter.context = context;
        this.computeConstants();
        
    }
    
 
    public BigDecimal getDecimalOne(){
        return this.decimalOne;    
    }
    
    public BigDecimal getDecimalZero(){
        return this.decimalZero;
    }
    
}
