/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.math.BigInteger;

/**
 *
 * @author fbio
 */
public class MarkovStatePrivateVariable extends MarkovStateVariable{

    Interval interval;
    boolean secret = false;
    //private static final Contexter contexter = Contexter.getInstance();
  
    
    public MarkovStatePrivateVariable(String name, String interval, int length) {
        super(name, true, length);
        this.interval = new Interval(interval);
    }
    
    public MarkovStatePrivateVariable(String name, String interval, Integer length) {
        super(name, true, length);
        this.interval = new Interval(interval);
    }
    
    public MarkovStatePrivateVariable(String name, Interval interval, int length) {
        super(name, true, length);
        this.interval = interval.clone();
    }
    
    public MarkovStatePrivateVariable(String name, Interval interval, Integer length) {
        super(name, true, length);
        this.interval = interval.clone();
    }

    public Interval getInterval() {
        return interval;
    }

    public void setInterval(String interv) {
        this.interval = new Interval(interv);
    }

    public void setInterval(Interval interv) {
        this.interval = interv;
    }
    
    public boolean isSecret() {
        return secret;
    }

    public void setSecret(boolean isSecret) {
        this.secret = isSecret;
    }
    
    
    
    @Override
    public String toString(){
        String toreturn;
        /*if (secret) {
            toreturn = this.getName().toUpperCase();
        }else{
            toreturn = this.getName().toLowerCase();
        }
        toreturn = toreturn + "="+this.interval;*/
        toreturn = this.getName()+"="+this.interval.toString();
        return toreturn;
    }
    
    public BigInteger getIntervalSize(){
       return this.interval.getSize();
       
               
    }
       
    private static void log(Object aObject, int level){
        Logger logger = Logger.getInstance();
        logger.log(String.valueOf(aObject),level);
    }
}
