/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.util.HashSet;
import java.util.Iterator;

/**
 *
 * @author fbio
 */
class MarkovStateSet extends HashSet<MarkovChainState> implements Cloneable{

    public boolean hasStateWithName(String name) {
        for (MarkovChainState s : this) {
            if (s.getName().equals(name)){
                return true;
            }
        }
        return false;
    }

    public MarkovChainState getStateByName(String name){
        Iterator<MarkovChainState> iter = this.iterator();
        MarkovChainState toreturn = null;
        while (iter.hasNext()){
            MarkovChainState s =iter.next();
            if (s.getName().equals(name)){
                toreturn = s;
                break;
            }
        }
        return toreturn;
    }

    MarkovChainState getStateByVariableIdentifier(String identifier) {
       Iterator<MarkovChainState> iter = this.iterator();
        MarkovChainState toreturn = null;
        while (iter.hasNext()){
            MarkovChainState s =iter.next();
            if (s.getVariables().getIdentifier().equals(identifier)){
                toreturn = s;
                break;
            }
        }
        return toreturn;
    }
   
    
    @Override
    public MarkovStateSet clone(){
        MarkovStateSet newset =(MarkovStateSet) super.clone();
        
        return newset;
    }
    
}
