/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.util.Comparator;

/**
 *
 * @author fbio
 */
abstract class MarkovStateVariable {
    private String name="";
    private boolean isPrivate = false;
    private Integer length = 1;

    public MarkovStateVariable(String name, boolean isPrivate, int length) {
        this.name = name;
        this.isPrivate = isPrivate;
        this.length = length;
    }
    
    public MarkovStateVariable(String name, boolean isPrivate, Integer length) {
        this.name = name;
        this.isPrivate = isPrivate;
        this.length = length;
    }

    public boolean isPrivate() {
        return isPrivate;
    }

    public void setPrivacy(boolean isPrivate) {
        this.isPrivate = isPrivate;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }
    
    
    
    public Integer getLength() {
        return length;
    }

    public void setLength(Integer length) {
        this.length = length;
    }
   
    @Override
    public String toString(){
        return name;      
    }
}

class MarkovStateVariableComparator implements Comparator<MarkovStateVariable>{

    @Override
    public int compare(MarkovStateVariable o1, MarkovStateVariable o2) {
        String ident1= o1.toString();
        String ident2 = o2.toString();
        return ident1.compareTo(ident2);
    }
    
}
