/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012-2014 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/

package quail;

import java.util.regex.Pattern;

/**
 *
 * @author fbio
 */
public class QuailPatterns {
     private static final String regex_constant_long = "[-+]?[0-9]+";
  private static final String regex_constant_double = "[-+]?[0-9]*\\.?[0-9]+";
  private static final String regex_intervals ="[\\["+regex_constant_long+","+regex_constant_long+"\\]]+";
  private static final String regex_variable = "[a-z][a-zA-Z0-9|\\_]*[\\[?\\([a-z][a-zA-Z0-9|\\_]?\\)\\]]*";
  private static final String regex_comparison_operator = "==|!=|>|<|>=|<=";
  private static final String regex_arithmetic_operator = "\\+|\\-|\\/|\\*|\\^|\\%";
  private static final String regex_boolean_operator = "\\|\\||\\&\\&";
  
  private static final String regex_operator = regex_comparison_operator+"|"+regex_arithmetic_operator+"|"+regex_boolean_operator;
  private static final String regex_expression_constant =  "("+regex_constant_long+")";
  public static final Pattern pattern_expression_constant = Pattern.compile( regex_expression_constant );
  private static final String regex_expression_variable =  "("+regex_variable+")";
  public static final Pattern pattern_expression_variable = Pattern.compile( regex_expression_variable );
  private static final String regex_negated_expression = "(\\![^;]+)";
  public static final Pattern pattern_negated_expression = Pattern.compile( regex_negated_expression );
  private static final String regex_expression_operation =  "("+regex_constant_long+")("+regex_operator+")("+regex_constant_long+")";
  public static final Pattern pattern_expression_operation = Pattern.compile( regex_expression_operation );
  
  private static final String regex_guard =  "\\(\\(("+regex_variable+"|"+regex_constant_long+")\\)("+regex_comparison_operator+")([^;]+)\\)";
  public static final Pattern pattern_guard = Pattern.compile( regex_guard );
  
  
  private static final String regex_public = "public int("+regex_constant_long+") ("+regex_variable+");";
  public static final Pattern pattern_public = Pattern.compile( regex_public );
  private static final String regex_public_with_assignment = "public int("+regex_constant_long+") ("+regex_variable+"):=("+regex_constant_long+");";
  public static final Pattern pattern_public_with_assignment = Pattern.compile( regex_public_with_assignment );
  private static final String regex_observable = "observable int("+regex_constant_long+") ("+regex_variable+");";
  public static final Pattern pattern_observable = Pattern.compile( regex_observable );
  private static final String regex_observable_with_assignment = "observable int("+regex_constant_long+") ("+regex_variable+"):=("+regex_constant_long+");";
  public static final Pattern pattern_observable_with_assignment = Pattern.compile( regex_observable_with_assignment );
  private static final String regex_goto = "goto ("+regex_constant_long+");";
  public static final Pattern pattern_goto = Pattern.compile( regex_goto );
  private static final String regex_private = "private int("+regex_constant_long+") ("+regex_variable+");";
  public static final Pattern pattern_private = Pattern.compile( regex_private );
  private static final String regex_private_with_assignment = "private int("+regex_constant_long+") ("+regex_variable+"):=("+regex_intervals+");";
  public static final Pattern pattern_private_with_assignment = Pattern.compile( regex_private_with_assignment );
  private static final String regex_secret = "secret int("+regex_constant_long+") ("+regex_variable+");";
  public static final Pattern pattern_secret = Pattern.compile( regex_secret );
  private static final String regex_secret_with_assignment = "secret int("+regex_constant_long+") ("+regex_variable+"):=("+regex_intervals+");";
  public static final Pattern pattern_secret_with_assignment = Pattern.compile( regex_secret_with_assignment );
  private static final String regex_return = "return;";
  public static final Pattern pattern_return = Pattern.compile( regex_return );
  private static final String regex_skip = "skip;";
  public static final Pattern pattern_skip = Pattern.compile( regex_skip );
  private static final String regex_assign = "assign \\(("+regex_variable+")\\):=(\\(.+\\));";
  public static final Pattern pattern_assign = Pattern.compile( regex_assign );
  private static final String regex_random = "random \\(("+regex_variable+")\\):=random\\((\\(.+\\))\\,(\\(.+\\))\\);";
  public static final Pattern pattern_random = Pattern.compile( regex_random );
  private static final String regex_randombit = "random \\(("+regex_variable+")\\):=randombit\\(("+regex_constant_double+")\\);";
  public static final Pattern pattern_randombit = Pattern.compile( regex_randombit );
  private static final String regex_if = "if (\\(.+\\))";
  public static final Pattern pattern_if = Pattern.compile( regex_if );
  private static final String regex_then = "then goto ("+regex_constant_long+");";
  public static final Pattern pattern_then = Pattern.compile( regex_then );
  private static final String regex_else = "else goto ("+regex_constant_long+");";
  public static final Pattern pattern_else = Pattern.compile( regex_else );
  private static final String regex_free = "free ("+regex_variable+");";
  public static final Pattern pattern_free = Pattern.compile( regex_free );
  private static final String regex_single_guard =  "\\(\\(("+regex_variable+"|"+regex_constant_long+")\\)("+regex_comparison_operator+")(\\([^;]+)\\)";
  public static final Pattern pattern_single_guard = Pattern.compile( regex_single_guard );
}
