/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package quail;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 *
 * @author fbio
 */
public class RunnableSingleExplorer implements Runnable{
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
    

    private ArrayList<String> codelines = null;
    private final ConcurrentSimplePosteriorDistributionManager postdistman;
    private final ThreadPoolExecutor tpoolexplorer;
    private final AtomicInteger count;
    private MarkovChainState currstate;
    private boolean terminated;
    private boolean parsingerror;
    
    
    RunnableSingleExplorer(MarkovChainState start, ArrayList<String> lines, ConcurrentSimplePosteriorDistributionManager postman, ThreadPoolExecutor tpool,AtomicInteger cont) {
        currstate = start;
        codelines = lines;
        postdistman = postman;
        tpoolexplorer = tpool;
        count = cont;
        
    }

    
    
    @Override
    public void run(){
        //initializing single exploration
        
        
     
        log("Starting exploration in state "+currstate.printNice()+" having probability "+currstate.getProbability().toEngineeringString(),3);
        /*if (initialstate == null){
            currstate = new MarkovChainState("current",contexter.getDecimalOne());
            MarkovStatePublicVariable programcounter = new MarkovStatePublicVariable("pc",BigInteger.ONE,Integer.valueOf(32));
            currstate.addVariable(programcounter);        
        }else{
            try {
                currstate =initialstate.clone();
            } catch (CloneNotSupportedException ex) {
                java.util.logging.Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, "Unable to clone state"+currstate.getName(), ex);
                System.exit(1);
            }
        }*/
        
        terminated = false;
        parsingerror = false;
        while (!terminated){
          BigInteger pc = currstate.getPublicVariableValue("pc");
          if (pc.compareTo(BigInteger.valueOf(-1)) ==0){
              log("ERROR: found state with no program counter information; exiting explorer",1);
              break;
          }
          if (currstate.getProbability().compareTo(contexter.getDecimalZero())==0){
              log("ERROR: current state has probability zero",1);
              System.exit(1);
          }
          String line = codelines.get(pc.intValue()); 
          log("",4);
          if (logger.getVerbosity()>=4){
              log("Exploring line "+pc+": "+line+" in state "+currstate.printNice(),4);
          }
          
          if (line.startsWith("private")){
              explorePrivateLine(line,pc,false);
          }else if (line.startsWith("public")){
              explorePublicLine(line,pc,false);
          }else if (line.startsWith("secret")){
              
              explorePrivateLine(line,pc,true);
          }else if (line.startsWith("observable")){
              explorePublicLine(line,pc,true);
          }else if (line.startsWith("assign")){
              exploreAssignLine(line,pc);
          }else if (line.startsWith("skip")){
              exploreSkipLine(line,pc);
          }else if (line.startsWith("random")){
              exploreRandomLine(line,pc);
          }else if (line.startsWith("goto")){
              exploreGotoLine(line,pc);
          }else if (line.startsWith("free")){
              exploreFreeLine(line,pc);
          }else if (line.startsWith("return")){
              exploreReturnLine(line,pc);
          }else if (line.startsWith("if")){
              exploreIfLine(line,codelines.get(pc.add(BigInteger.ONE).intValue()),codelines.get(pc.add(new BigInteger("2")).intValue()),pc);
          }else {
              log("ERROR: line "+pc.toString()+" unrecognized, this should never happen; exiting parser",1);
              parsingerror = true;
              break;
          }
          if (parsingerror){
              log("ERROR: parsing error; exiting parser",1);
              System.exit(1);
          }  
        }
        log("Found final state "+currstate.printNice(),3);
       
   
        
        int c = count.decrementAndGet();
        log("Decreasing count to "+c,3);
        if (c== 0){
            tpoolexplorer.shutdown();
            log("I am the last thread, terminating the pool. Active count = "+tpoolexplorer.getActiveCount(),2);
        } 
        postdistman.inferInformation(currstate);
    }
    
    private void explorePrivateLine(String aLine, BigInteger pc, boolean isSecret) {
        String varname;
       Integer varlength;
        String varinterval;
    
    
    /*
     * String input = "1 fish 2 fish red fish blue fish";
     Scanner s = new Scanner(input);
     s.findInLine("(\\d+) fish (\\d+) fish (\\w+) fish (\\w+)");
     MatchResult result = s.match();
     for (int i=1; i<=result.groupCount(); i++)
         System.out.println(result.group(i);
     s.close(); 
     */
    
      Pattern pattern;
       if (isSecret){
          pattern = QuailPatterns.pattern_secret;
       }else{
          pattern = QuailPatterns.pattern_private;
      }
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+pattern.pattern(),4);
    }
    
    Matcher matcher = pattern.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable length: " + matcher.group(1),4);
      log("Variable name: " + matcher.group(2),4);
      varlength = Integer.valueOf(matcher.group(1));
      varname = matcher.group(2);
      BigInteger interend = BigInteger.valueOf(2).pow(varlength.intValue()).subtract(BigInteger.ONE);
      varinterval="[0,"+interend+"]";
    }else{
        if (isSecret){
        pattern = QuailPatterns.pattern_secret_with_assignment ;
    }else{
        pattern = QuailPatterns.pattern_private_with_assignment ;
    }
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+pattern.pattern(),4);
    }
    
    matcher = pattern.matcher(aLine );
    if ( matcher.find() ) {
          log("Num groups: " + matcher.groupCount(),4);
          log("Variable length: " + matcher.group(1),4);
          log("Variable name: " + matcher.group(2),4);
          log("Assigned interval: "+matcher.group(3),4);
          varlength = Integer.valueOf(matcher.group(1));
          varname = matcher.group(2);
          varinterval=matcher.group(3);
        }else{
            log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
            parsingerror = true;
            return;
        }
        
    }
    
    
    
    
    MarkovStatePrivateVariable var = new MarkovStatePrivateVariable(varname,varinterval,varlength);
    var.setSecret(isSecret);
    currstate.addVariable(var);

    if (logger.getVerbosity()>=3){
        if (isSecret){
            log("Added secret variable "+var.toString()+" with "+var.getIntervalSize()+" possible values "+varinterval+" to state "+currstate.getName(),3);        
        }else{
            log("Added private variable "+var.toString()+" with "+var.getIntervalSize()+" possible values "+varinterval+" to state "+currstate.getName(),3);
        }
    }
    
    
      
    //
       
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
                
    }

    private void explorePublicLine(String aLine, BigInteger pc, boolean isObservable) {
        if (isObservable){
        log("Parsing observable line: "+aLine,4);
    }else{
        log("Parsing public line: "+aLine,4);
    }
       
       
    //Scanner scanner = new Scanner(aLine);
    String varname;
    Integer varlength;
    BigInteger varvalue;
    
    
    /*
     * String input = "1 fish 2 fish red fish blue fish";
     Scanner s = new Scanner(input);
     s.findInLine("(\\d+) fish (\\d+) fish (\\w+) fish (\\w+)");
     MatchResult result = s.match();
     for (int i=1; i<=result.groupCount(); i++)
         System.out.println(result.group(i);
     s.close(); 
     */
    /*
    scanner.findInLine("public int[\d+] \w+;");
    MatchResult result = scanner.match();*/
    Pattern pattern;
    if (isObservable){
        pattern = QuailPatterns.pattern_observable ;
    }else{
        pattern = QuailPatterns.pattern_public ;
    }
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+pattern.pattern(),4);
    }
    
    Matcher matcher = pattern.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable length: " + matcher.group(1),4);
      log("Variable name: " + matcher.group(2),4);
      varlength = Integer.valueOf(matcher.group(1));
      varname = matcher.group(2);
      varvalue = BigInteger.ZERO;
    
    }else{
        if (isObservable){
            pattern = QuailPatterns.pattern_observable_with_assignment ;
        }else{
            pattern = QuailPatterns.pattern_public_with_assignment ;
        }
        if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+pattern.pattern(),4);
        }
        
        matcher = pattern.matcher(aLine );
        if ( matcher.find() ) {
          log("Num groups: " + matcher.groupCount(),4);
          log("Variable length: " + matcher.group(1),4);
          log("Variable name: " + matcher.group(2),4);
          log("Assigned value: "+matcher.group(3),4);
          varlength = Integer.valueOf(matcher.group(1));
          varname = matcher.group(2);
          try{
          varvalue = evaluateExpression(matcher.group(3),currstate.getVariables());
          } catch (ParseException e) {
              log("ERROR: unable to evaluate expression "+matcher.group(3)+" in state "+currstate.printNice(),1);
            
              parsingerror = true;
              return;
          }
    
        }else{
            log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
     
            parsingerror = true;
            return;
        }
    }
    

    MarkovStatePublicVariable var = new MarkovStatePublicVariable(varname,varvalue,varlength);
    var.setObservable(isObservable);
    currstate.addVariable(var);

    if (logger.getVerbosity()>=3){
        if (isObservable){
            log("Added observable variable "+var.toString()+" to state "+currstate.getName(),3);
        }else{
            log("Added public variable "+var.toString()+" to state "+currstate.getName(),3);
        }
    }
    
    
  
    //
   
    
   
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
  
  
   
    }
    
     private BigInteger evaluateExpression(String expr, VariableList vars) throws ParseException{
       boolean found = false;
       while (found == false){
           // look for a closed parenthesis
           int clospos =-1;
           for (int i=0;i<expr.length();i++){
               if (expr.charAt(i)==')'){
                   clospos = i;
                   break;
               }
           }
       
           // if there are no closed parenthesis the expression is a constant; return its value
           if (clospos==-1){
               return this.evaluateInnerExpression(expr, vars);
           
           }else{
               //else find the last opened parenthesis before the closed one
               String smallexpr = expr.substring(0, clospos+1);
               log ("Found closed parenthesis: "+smallexpr,4);
               int openpos = smallexpr.lastIndexOf('(');
               smallexpr = smallexpr.substring(openpos);
               log("Found inner expression "+smallexpr,4);
               // found an inner expression; evaluate it (without the parenthesis)
               BigInteger innerval = new BigInteger("0");
               try{
                   innerval = this.evaluateInnerExpression(smallexpr.substring(1, smallexpr.length()-1), vars);
               }catch (ParseException e){
                   throw e;
               }
               
               String smallexprval = ""+innerval;
               expr = expr.replace(smallexpr, smallexprval);
               log("Replaced inner expression "+smallexpr+" with its value "+smallexprval+" in the string; now the string to evaluate is "+expr,4);
           
           }
       }
       
       
       return BigInteger.valueOf(-1);
   }

    private BigInteger evaluateInnerExpression (String expr, VariableList vars)throws ParseException {
        log("Evaluating inner expression "+expr+" with variables "+vars.toString(),4);
        if (expr.startsWith("(")){
            expr = expr.substring(1,expr.lastIndexOf(")"));
            log("Removing parenthesis; the expression is now "+expr,4);
        }
        if (logger.getVerbosity()>=4){
            log("The regular expression for a constant is "+QuailPatterns.pattern_expression_constant.pattern(),4);
        }
        
        Matcher matcher = QuailPatterns.pattern_expression_constant.matcher(expr );
        if ( matcher.matches() ) {
            log("It is the constant " + matcher.group(1)+"; returning it.",4);
            return new BigInteger(matcher.group(1));
        }else{
            log("It is not a constant, checking if it is a variable",4);
            if (logger.getVerbosity()>=4){
                log("The regular expression for a variable is "+QuailPatterns.pattern_expression_variable.pattern(),4);
            }
            
            matcher = QuailPatterns.pattern_expression_variable.matcher(expr );
            if ( matcher.matches() ) {
                log("It is the variable " + matcher.group(1)+"; checking its value...",4);
                String varname = this.getVariableName(matcher.group(1), vars);
                MarkovStateVariable var = vars.getVariable(varname);
                if (var==null){
                    log("ERROR: variable "+varname+" not found.",4);
                    throw new ParseException("ERROR: variable "+varname+" not found.",0);
                }else if (var.isPrivate()){
                    log("ERROR: variable "+varname+" is private and thus has no defined value.",4);
                    throw new ParseException("ERROR: variable "+varname+" is private and thus has no defined value.",0);
                }else{
                    BigInteger val = ((MarkovStatePublicVariable)var).getVal();
                    log("The value is "+val+"; returning it.",4);
                    return val;
                }
                
                
            }else{
                log("It is not a variable, checking if it is an operation",4);
                if (logger.getVerbosity()>=4){
                    log("The regular expression for an operation is "+QuailPatterns.pattern_expression_operation.pattern(),4);
                }
                
                matcher = QuailPatterns.pattern_expression_operation.matcher(expr );
                if ( matcher.matches() ) {
                    String op1 = matcher.group(1);
                    String operator = matcher.group(2);
                    String op2 = matcher.group(3);
                    log("It is the operation "+op1+" "+operator+" "+op2+"; evaluating it...",4);
                    switch (operator) {
                        case "+":
{
                                log("It is a sum; evaluating the operands",4);
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.add(val2);
                                log("The sum result is "+result,4);
                                return result;
                            }
                        case "-":
{
                                log("It is a subtraction; evaluating the operands",4);
                             
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.subtract(val2);
                                log("The subtraction result is "+result,4);
                                return result;
                            }
                        case "/":
{
                                log("It is a division; evaluating the operands",4);
                                
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.divide(val2);
                                log("The division result is "+result,4);
                                return result;
                            }
                        case "*":
{
                                log("It is a product; evaluating the operands",4);
                               
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.multiply(val2);
                                log("The product result is "+result,4);
                                return result;
                            }
                        case "%":
{
                                log("It is a modulo; evaluating the operands",4);
                               
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.mod(val2);
                                log("The modulo result is "+result,4);
                                return result;
                            }
                        case "&&":
{
                                log("It is a logical AND; evaluating the operands",4);
                               
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result;
                                if (val1.compareTo(BigInteger.ZERO)==0 && val2.compareTo(BigInteger.ZERO)==0){
                                    result = BigInteger.ZERO;
                                }else{
                                    result = BigInteger.ONE;
                                }
                                log("The AND result is "+result,4);
                                return result;
                            }
                        case "||":
{
                                log("It is a logical OR; evaluating the operands",4);
                               
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result;
                                if (val1.compareTo(BigInteger.ZERO)==0 || val2.compareTo(BigInteger.ZERO)==0){
                                    result = BigInteger.ZERO;
                                }else{
                                    result = BigInteger.ONE;
                                }
                                log("The OR result is "+result,4);
                                return result;
                            }
                        case "^":
{
                                log("It is a logical XOR; evaluating the operands",4);
                                
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result;
                                if ((val1.compareTo(BigInteger.ZERO)==0 && val2.compareTo(BigInteger.ZERO)==0) || (val1.compareTo(BigInteger.ZERO)!=0 && val2.compareTo(BigInteger.ZERO)!=0)){
                                    result = BigInteger.ZERO;
                                }else{
                                    result = BigInteger.ONE;
                                }
                                log("The XOR result is "+result,4);
                                return result;
                            }
                    }
                    log("ERROR: operator not recognized.",1);
                    throw new ParseException("ERROR: operator not recognized;",0);
                }else {
                 log("It is not an operation, checking if it is a negation",4);
                if (logger.getVerbosity()>=4){
                    log("The regular expression for a negation is "+QuailPatterns.pattern_negated_expression.pattern(),4);
                
                }
                matcher = QuailPatterns.pattern_negated_expression.matcher(expr );
                if ( matcher.matches() ) {
                    String negexpr = matcher.group(1);
                   
                    log("It is the negated operation "+negexpr+"; negating it...",4);
                    BigInteger negval = new BigInteger(negexpr.substring(1));
                    
                    if (negval.compareTo(BigInteger.ZERO)==0){
                        negval = BigInteger.ONE;
                    }else{
                        negval = BigInteger.ZERO;
                    }
                    return negval;
                }else {
                    log("ERROR: string "+expr+" does not match any known inner expression.",1);
                    throw new ParseException("ERROR: string "+expr+" does not match any known inner expression.",0);
                }
            }
            }


        }
    }

    private void exploreRandomLine(String line, BigInteger pc){
        String varname;
        BigInteger newpc = pc.add(BigInteger.ONE);
        if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+QuailPatterns.pattern_randombit.pattern(),4);
        }
        
        Matcher matcher = QuailPatterns.pattern_randombit.matcher(line );
        if ( matcher.find() ) {
          log("Num groups: " + matcher.groupCount(),4);
          log("Variable name: " + matcher.group(1),4);
          log("Probability of being 0: " + matcher.group(2),4);
          varname = this.getVariableName(matcher.group(1),currstate.getVariables());
          BigDecimal probofzero = new BigDecimal(matcher.group(2),contexter.getContext());
          
          if (probofzero.compareTo(contexter.getDecimalOne())>0){
                log("Found a probability greater than 1; using 1 instead",4);
                probofzero = contexter.getDecimalOne();
            } else if (probofzero.compareTo(contexter.getDecimalZero())<0){
                log("Found a probability lesser than 0; using 0 instead",4);
                probofzero = contexter.getDecimalZero();
            }


         
           
          log("Assiging to public variable "+varname+" value 0 with probability "+probofzero+" in state "+currstate.getName(),3);
          if (probofzero.compareTo(contexter.getDecimalOne())==0){
              currstate.setPublicVariable(varname,BigInteger.ZERO);
              log("The value of "+varname+" is 0 with probability 1",3);
          }else if (probofzero.compareTo(contexter.getDecimalZero())==0){
              currstate.setPublicVariable(varname,BigInteger.ONE);
              log("The value of "+varname+" is 1 with probability 1",3);
          }else{
               log("Exploring the case in which variable "+varname+" is zero and putting the case in which it is 1 on the stack for future exploration",3);
              
               
               currstate.setPublicVariable(varname,BigInteger.ZERO);
               
               currstate.setPublicVariable("pc", newpc);
               
               MarkovChainState otherstate = null;
                  try {
                        otherstate = currstate.clone();
                  } catch (CloneNotSupportedException ex) {
                      java.util.logging.Logger.getLogger(RunnableSingleExplorer.class.getName()).log(Level.SEVERE, "Unable to clone state "+currstate.getName(), ex);
                      System.exit(1);
               }
               otherstate.setPublicVariable(varname,BigInteger.ONE);
               otherstate.multiplyProbabilityBy(contexter.getDecimalOne().subtract(probofzero,contexter.getContext()));
               //toexplore.add(otherstate);
               int c = count.incrementAndGet();
               log("Increasing count to "+c,3);
               tpoolexplorer.submit(new RunnableSingleExplorer(otherstate,codelines,postdistman,tpoolexplorer,count));
               
               currstate.multiplyProbabilityBy(probofzero);
               /*BigDecimal onehalf = new BigDecimal("0.5",contexter.getContext());
               if (toss.compareTo(onehalf)>=0){
                  // the bit is 0
                  currstate.setPublicVariable(varname,BigInteger.ZERO);
                  currstate.multiplyProbabilityBy(probofzero);
                  log("Randomly chose to assign value 0 to variable "+varname,3);
              }else{
                  // the bit is 1
                  currstate.setPublicVariable(varname,BigInteger.ONE);
                  currstate.multiplyProbabilityBy(contexter.getDecimalOne().subtract(probofzero, contexter.getContext()));
                  log("Randomly chose value 1 to variable "+varname,3);
              }*/
          }

      
           currstate.setPublicVariable("pc", newpc);
          
          
        }else{
            if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+QuailPatterns.pattern_random.pattern(),4);
            }

            matcher = QuailPatterns.pattern_random.matcher(line );
            if ( matcher.find() ) {
                log("Num groups: " + matcher.groupCount(),4);
          log("Variable name: " + matcher.group(1),4);
          log("Lower bound: " + matcher.group(2),4);
          log("Upper bound: " + matcher.group(3),4);
          varname = this.getVariableName(matcher.group(1),currstate.getVariables());

        BigInteger lowbound = new BigInteger("0");
        try{
            lowbound = evaluateExpression(matcher.group(2),currstate.getVariables());
        }catch (ParseException e){
            log(e.getMessage(),1);
        }

 
        BigInteger uppbound = new BigInteger("0");
        try{
            uppbound = evaluateExpression(matcher.group(3),currstate.getVariables());
        }catch (ParseException e){
            log(e.getMessage(),1);
        }
           
          if (lowbound.compareTo(uppbound)>0){
              log("ERROR: lower bound greater than upper bound?",2);
          
              parsingerror = true;
              return;
          }
          
          BigInteger diff = uppbound.subtract(lowbound).add(BigInteger.ONE);
          if (diff.compareTo(BigInteger.ONE)==0){
            //VariableList vars =currstate.getVariables().clone();
            currstate.setPublicVariable(varname, uppbound);
            log("Assigned to public variable "+varname+" value "+uppbound.toString()+" with probability 1 in state "+currstate.getName(),3);
           


            currstate.setPublicVariable("pc", newpc);
           
            
            
          }else{
              BigDecimal probofeachvalue = new BigDecimal(diff);
              //BigDecimal toss = new BigDecimal(Math.random()).multiply(decdiff);
              
              
              currstate.setPublicVariable(varname, lowbound);
              currstate.multiplyProbabilityBy(contexter.getDecimalOne().divide(probofeachvalue, contexter.getContext()));
              currstate.setPublicVariable("pc", newpc);
              
              log("Next state explored will be "+currstate.printNice()+" with probability "+currstate.getProbability().toEngineeringString(),3);
              
              BigInteger value = lowbound.add(BigInteger.ONE);
              while (value.compareTo(uppbound)<=0){
                  MarkovChainState otherstate = null;
                  try {
                        otherstate = currstate.clone();
                  } catch (CloneNotSupportedException ex) {
                      java.util.logging.Logger.getLogger(RunnableSingleExplorer.class.getName()).log(Level.SEVERE, "Unable to clone state "+currstate.getName(), ex);
                      System.exit(1);
                  }
                  otherstate.setPublicVariable(varname, value);
                  log("Added state "+otherstate.printNice()+" with probability "+otherstate.getProbability().toEngineeringString()+" to the stack",3);
              
                  //toexplore.add(otherstate);
                  int c = count.incrementAndGet();
                  log("Increasing count to "+c,3);
                  tpoolexplorer.submit(new RunnableSingleExplorer(otherstate,codelines,postdistman,tpoolexplorer,count));
                  value = value.add(BigInteger.ONE);
              }
              
              log("Exploring the case in which variable "+varname+" is "+lowbound+" and putting the other cases on the stack",3);

      

    
          }
        
            }else{
                log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);

             
                parsingerror = true;
            }
                
        }
        
    }
    private String getVariableName(String varname, VariableList variables) {
        String toreturn =varname;
        if (varname.contains("[")){
            String temp = varname.substring(varname.indexOf('['));
            while (temp.contains("[")){
                int closedpar = temp.indexOf(']');
                String index = temp.substring(1, closedpar);
                log("Evaluating index "+index,4);
                index = this.getVariableName(index, variables);
                BigInteger val = new BigInteger("0");
                try {
                    val = this.evaluateInnerExpression(index, variables);
                } catch (ParseException e){
                    log(e.getMessage(),1);
                }
                toreturn = toreturn.replace("["+index+"]", "["+val.toString()+"]");
                temp = temp.substring(closedpar+1);
            }
                    
            
        }
        
        
        
        
        return toreturn;
    }

    private void exploreAssignLine(String aLine, BigInteger pc) {
        String varname;
    String varexpression;
    BigInteger varvalue;
    
    
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_assign.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_assign.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable name: " + matcher.group(1),4);
      log("Variable expression: " + matcher.group(2),4);
      varname = this.getVariableName(matcher.group(1),currstate.getVariables());
      varexpression = matcher.group(2);
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        
        parsingerror = true;
        return;
    }
    VariableList vars =currstate.getVariables().clone();
    try{
        varvalue = evaluateExpression(varexpression,vars);
    }catch (ParseException e){
        log("ERROR: unable to evaluate expression "+varexpression+" in state "+currstate.printNice(),1);
     
        parsingerror = true;
        return;
    }
    
    log("Assigned to public or observable variable "+varname+" value "+varvalue+" in state "+currstate.getName(),3);
    
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
    currstate.setPublicVariable(varname, varvalue);
    
    }

    private void exploreIfLine(String ifline, String thenline, String elseline, BigInteger pc) {

    
    String guard;
    BigInteger gotothen,gotoelse;
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_if.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_if.matcher(ifline );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("If guard: " + matcher.group(1),4);
      guard = matcher.group(1);
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
       
        parsingerror = true;
        return;
    }
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_then.pattern(),4);
    }
    
    matcher = QuailPatterns.pattern_then.matcher(thenline );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Then destination: " + matcher.group(1),4);
      gotothen = new BigInteger(matcher.group(1));
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc").add(BigInteger.ONE) +" does not match regular expression.",1);
      
        parsingerror = true;
        return;
    }
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_else.pattern(),4);
    }
    
    matcher = QuailPatterns.pattern_else.matcher(elseline );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Else destination: " + matcher.group(1),4);
      gotoelse = new BigInteger(matcher.group(1));
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc").add(BigInteger.ONE).add(BigInteger.ONE) +" does not match regular expression.",1);
       
        parsingerror = true;
        return;
    }
    
    log("Exploring if statement with guard "+guard+", then line "+gotothen+" and else line "+gotoelse+" to state "+currstate.printNice(),3);
    
    
    String varname;
    String operator;
    String expression;
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_guard.pattern(),4);
    }
    
    matcher = QuailPatterns.pattern_guard.matcher(guard );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable name: " + matcher.group(1),4);
      log("Operator: " + matcher.group(2),4);
      log("Guard expression: "+matcher.group(3),4);
      varname = this.getVariableName(matcher.group(1),currstate.getVariables());
      operator = matcher.group(2);
      expression = matcher.group(3);
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        
        parsingerror = true;
        return;
    }
   
    // evaluate the guard expression
        BigInteger expval = new BigInteger("0");
        try{
            expval = this.evaluateExpression(expression, currstate.getVariables());
        }catch (ParseException e){
            log(e.getMessage(),1);
        }
        log("The expression evaluates to "+expval,4);
    
   boolean secretdependent = true;
   BigInteger argval = BigInteger.ZERO;
   log("The regular expression for a constant is "+QuailPatterns.pattern_expression_constant.pattern(),4);
   matcher = QuailPatterns.pattern_expression_constant.matcher(varname);
   if ( matcher.matches() ) {
       log("The argument is the constant " + matcher.group(1),4);
       argval = new BigInteger(matcher.group(1));
       secretdependent = false;
   }else{
    
        // obtain the variable
        MarkovStateVariable var = currstate.getVariables().getVariable(varname);
            
        // if the variable is private, the if is secret-dependent; if not take the value
        if (var.isPrivate()==false){
            MarkovStatePublicVariable pubvar = (MarkovStatePublicVariable)var;
            argval = pubvar.getVal();
            log("Variable "+pubvar.getName()+" is public and has value "+argval,4);
            secretdependent = false;
        }
        
    }
        
   if (secretdependent){
       VariableList thenvars =currstate.getVariables().clone();
       thenvars.setPublicVariable("pc", gotothen);
       VariableList elsevars =currstate.getVariables().clone();
       elsevars.setPublicVariable("pc", gotoelse);
    
       thenvars = processGuard(thenvars,varname,operator,expval,true);
       elsevars = processGuard(elsevars,varname,operator,expval,false);
       
    
       MarkovStatePrivateVariable hbefore =(MarkovStatePrivateVariable) currstate.getVariables().getVariable(varname);
       MarkovStatePrivateVariable hthen =(MarkovStatePrivateVariable) thenvars.getVariable(varname);
       MarkovStatePrivateVariable helse =(MarkovStatePrivateVariable) elsevars.getVariable(varname);
    
       BigDecimal thensize = new BigDecimal(hthen.getIntervalSize(),contexter.getContext());
       BigDecimal beforesize = new BigDecimal(hbefore.getIntervalSize(),contexter.getContext());
       BigDecimal thenprob = thensize.divide(beforesize,contexter.getContext());
      
    
       log("The guard is true with probability "+thenprob.toString(),3);
       if (thenprob.compareTo(contexter.getDecimalOne())==0){
           // the guard is certainly true
           currstate.setPublicVariable("pc",gotothen);
           currstate.setPrivateVariable(varname,hthen.getInterval().toString());
           log("The guard is always true; reducing variable "+varname+" to "+hthen.getInterval(),3);
       }else if (thenprob.compareTo(contexter.getDecimalZero())==0){
           // the guard is certainly false
           currstate.setPublicVariable("pc",gotoelse);
           currstate.setPrivateVariable(varname,helse.getInterval().toString());
           log("The guard is always false; reducing variable "+varname+" to "+helse.getInterval(),3);
       }else{
           // the guard is true with probability thenprob
           //log("Exploring the case in which the guard is true and variable "+varname+" is "+hthen.getInterval()+"; putting the case in which the guard is false and variable "+varname+" is "+helse.getInterval()+" on the stack",3);
           
           MarkovChainState otherstate = null;
              try {
                    otherstate = currstate.clone();
              } catch (CloneNotSupportedException ex) {
                  java.util.logging.Logger.getLogger(RunnableSingleExplorer.class.getName()).log(Level.SEVERE, "Unable to clone state "+currstate.getName(), ex);
                  System.exit(1);
           }
              
           currstate.setPublicVariable("pc",gotothen);
           currstate.setPrivateVariable(varname,hthen.getInterval().toString());
           currstate.multiplyProbabilityBy(thenprob);
           log("Next state explored will be "+currstate.printNice()+" with probability "+currstate.getProbability().toEngineeringString(),3);
           
           BigDecimal elseprob = contexter.getDecimalOne().subtract(thenprob,contexter.getContext());
           otherstate.setPublicVariable("pc",gotoelse);
           otherstate.setPrivateVariable(varname,helse.getInterval().toString());
           otherstate.multiplyProbabilityBy(elseprob);
           
           //log("thenprob is "+thenprob.toEngineeringString()+", elseprob is "+elseprob.toEngineeringString(),3);
           log("Put on the stack state "+otherstate.printNice()+" with probability "+otherstate.getProbability().toEngineeringString(),3);
           //toexplore.add(otherstate);
           int c = count.incrementAndGet();
           log("Increasing count to "+c,3);
           tpoolexplorer.submit(new RunnableSingleExplorer(otherstate,codelines,postdistman,tpoolexplorer,count));
       }
       
       
    
    
    

   }else{
       // the if does not depend on any private variables, and the value of the argument is known
       // since ifs don't have side effects, no variables except pc will be changed; we just need
       // to know if the next state will have the pc value of the then or of the else
       boolean guardistrue = false;
            switch (operator) {
                case "==":
                    if (argval.compareTo(expval)==0) {
                        guardistrue = true;
                    }
                    break;
                case "!=":
                    if (argval.compareTo(expval)!=0) {
                        guardistrue = true;
                    }
                    break;
                case ">=":
                    if (argval.compareTo(expval)>=0) {
                        guardistrue = true;
                    }
                    break;
                case "<=":
                    if (argval.compareTo(expval)<=0) {
                        guardistrue = true;
                    }
                    break;
                case "<":
                    if (argval.compareTo(expval)<0) {
                        guardistrue = true;
                    }
                    break;
                case ">":
                    if (argval.compareTo(expval)>0){
                        guardistrue = true;
                    }
                
                    break;
            }
       
       if (guardistrue) {
           currstate.setPublicVariable("pc", gotothen);
       }else{
           currstate.setPublicVariable("pc", gotoelse);
       }

       
   }
    
    }
    
    private VariableList processGuard(VariableList vars, String varname, String operator, BigInteger expval, boolean b) {
        if (logger.getVerbosity()>=4){
            log("Processing guard "+varname+operator+expval+" with variables "+vars.toString(),4);
        }

        // invert the operator if needed
        if (b==false){
            log("Inverting the operator as requested; original operator: "+operator,4);
            switch (operator) {
                case "==":
                    operator = "!=";
                    break;
                case "!=":
                    operator = "==";
                    break;
                case ">=":
                    operator = "<";
                    break;
                case "<=":
                    operator = ">";
                    break;
                case "<":
                    operator = ">=";
                    break;
                case ">":
                    operator = "<=";
                    break;
            }
            log ("New operator: "+operator,4);
        }
        
        // the guard is reduced to private variable -- operator -- constant; now we reduce the allowed intervals of the private variable accordingly
       MarkovStatePrivateVariable privvar = (MarkovStatePrivateVariable)vars.getVariable(varname);
       String intervals = privvar.getInterval().toString();
       String newintervals = "";
       
       while (intervals.compareTo("")!=0){
           // look for a closed parenthesis
           int clospos =-1;
           for (int i=0;i<intervals.length();i++){
               if (intervals.charAt(i)==']'){
                   clospos = i;
                   break;
               }
           }
       
           // if there are no closed parenthesis we are done; return
           if (clospos==-1){
               break;
           
           }else{
               //else find the last opened parenthesis before the closed one
               String oneinterval = intervals.substring(0, clospos+1);
               log ("Found closed parenthesis: "+oneinterval,4);
               int openpos = oneinterval.lastIndexOf('[');
               oneinterval = oneinterval.substring(openpos);
               log("Found single interval "+oneinterval,4);
               // found an inner expression; evaluate it (without the parenthesis)
               int commapos = oneinterval.lastIndexOf(',');
               String bottomstring = oneinterval.substring(1, commapos);
               log("Bottom string: "+bottomstring,4);
               BigInteger bottom = new BigInteger(bottomstring);
               String topstring = oneinterval.substring(commapos+1, oneinterval.length()-1);
               log("Top string: "+topstring,4);
               BigInteger top = new BigInteger(topstring);
               
               String onenewinterval = this.reduceInterval(bottom,top,operator,expval);
               newintervals += onenewinterval;
               
               if (intervals.lastIndexOf('[', oneinterval.length())==-1){
                   intervals = "";
               }else{
                   intervals = intervals.substring(oneinterval.length());
               }
               
           }
       }
       privvar.setInterval(newintervals);
       if (logger.getVerbosity()>=4){
           log("Setting intervals of variable "+privvar.getName()+" to "+privvar.getInterval(),4);
       }
        
        return vars;
    }
    
    private String reduceInterval(BigInteger bottom, BigInteger top, String operator, BigInteger expval) {
        log("Reducing inverval ["+bottom+","+top+"] assuming that "+operator+expval+" holds",4);
        String toreturn = "";
        switch (operator) {
            case "==":
                // processing equality
                // if the value is outside the range return empty; else return singleton interval
                if ((expval.compareTo(bottom)>=0)&&(expval.compareTo(top)<=0)){
                    toreturn = "["+expval+","+expval+"]";
                }else{
                    toreturn = "";
                }
                break;
            case "!=":
                // processing inequality
                // if the value is less than bottom or more than top, nothing to do
                if ((expval.compareTo(bottom)<0)||(expval.compareTo(top)>0)){
                    toreturn = "["+bottom+","+top+"]";
                }else if ((expval.compareTo(bottom)==0)&&(expval.compareTo(top)==0)){
                    // return empty string
                    toreturn = "";
                }else if (expval.compareTo(bottom)==0){
                    // cut the bottom value
                    toreturn = "["+(bottom.add(BigInteger.ONE))+","+top+"]";
                }else if (expval.compareTo(top)==0){
                    // cut the top value
                    toreturn = "["+bottom+","+(top.subtract(BigInteger.ONE))+"]";
                }else{
                    // the value is inside; we need to split the interval in two
                    toreturn = "["+bottom+","+(expval.subtract(BigInteger.ONE))+"]["+(expval.add(BigInteger.ONE))+","+top+"]";
                }
                break;
            case ">=":
                // processing greater or equal
                // if the value is lesser or equal to bottom, nothing to do
                if (expval.compareTo(bottom)<=0){
                    toreturn = "["+bottom+","+top+"]";
                }else if (expval.compareTo(top)>0){
                    // if the value is greater that top, return empty string
                    toreturn = "";
                }else {
                    // else move the bottom to expval
                    toreturn = "["+expval+","+top+"]";
                }
                break;
            case "<=":
                // processing lesser or equal
                // if the value is greater or equal to top, nothing to do
                if (expval.compareTo(top)>=0){
                    toreturn = "["+bottom+","+top+"]";
                }else if (expval.compareTo(bottom)<0){
                    // if the value is lesser that bottom, return empty string
                    toreturn = "";
                }else {
                    // else move the top to expval
                    toreturn = "["+bottom+","+expval+"]";
                }
                break;
            case "<":
                // processing lesser
                // if the value is greater than top, nothing to do
                if (expval.compareTo(top)>0){
                    toreturn = "["+bottom+","+top+"]";
                }else if (expval.compareTo(bottom)<=0){
                    // if the value is lesser or equal that bottom, return empty string
                    toreturn = "";
                }else {
                    // else move the top to expval-1
                    toreturn = "["+bottom+","+(expval.subtract(BigInteger.ONE))+"]";
                }
                break;
            case ">":
                // processing greater
                // if the value is lesser than bottom, nothing to do
                if (expval.compareTo(bottom)<0){
                    toreturn = "["+bottom+","+top+"]";
                }else if (expval.compareTo(top)>=0){
                    // if the value is greater or equal than top, return empty string
                    toreturn = "";
                }else {
                    // else move the bottom to expval+1
                    toreturn = "["+(expval.add(BigInteger.ONE))+","+top+"]";
                }
                break;
        }
        log("Reduced to "+toreturn,4);
        return toreturn;
    }

    private void exploreGotoLine(String aLine, BigInteger pc) {
        BigInteger gotoline;
    
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_goto.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_goto.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Goto line destination: " + matcher.group(1),4);
      gotoline = new BigInteger(matcher.group(1));
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
    
        parsingerror = true;
        return;
    }
   
    log("Adding goto transition from state "+currstate.getName()+" to line "+gotoline,3);
    
    
    
      
    BigInteger newpc = gotoline;
   
    currstate.setPublicVariable("pc", newpc);
 
 
   
    }
    
    private void exploreSkipLine(String aLine, BigInteger pc) {
   
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_skip.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_skip.matcher(aLine );
    if ( matcher.find()==false ) {
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        parsingerror = true;
        return;
    }
    log("Adding skip transition in state "+currstate.getName(),3);
       
        
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
                
   }
    

    private void exploreReturnLine(String aLine, BigInteger pc) {
        if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_return.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_return.matcher(aLine );
    if ( matcher.find()==false ) {
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
       
        parsingerror = true;
        return;
    }
      log("Found return statement. Terminating.",3);
      this.terminated = true;
   
    }
    
    private void exploreFreeLine(String aLine, BigInteger pc) {
        
       
    String varname;
        
    
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_free.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_free.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable name: " + matcher.group(1),4);
      varname = this.getVariableName(matcher.group(1),currstate.getVariables());
      
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
       
        parsingerror = true;
        return;
    }
  
    
    
    log("Freed variable "+varname+" in state "+currstate.getName(),3);
    
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
    currstate.deleteVariableByName(varname);
    
  
    }
    
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }
    
}
